#include "defo.h"
#include "find-symmetry.h"

MyInterval normEucl2D(MyInterval x, MyInterval y)
/** Returns Euclidean norm of vector (x,y) */
{
    return sqrt((power(x, 2) + power(y, 2)));
}

// 2D
void writesymmetryFromPerm2D(int permutation[], int N, vector<vector<int>> &cycles, int noC, bool sg, MyMatrix R)
// the function writes to cout the data about symmetry
{
    // sg --- sign of det of matrix R:
    // sg = TRUE <==> det(R) = 1 <==> R+
    // sg = FALSE <==> det(R) = -1 <==> R-
    // noC -- number of cycles in vector 'cycles'
    int oneElementC = 0;  // the noumber of 1-element cycles
                          //    int twoElementC = 0;  // the noumber of 2-element cycles
    int manyElementC = 0; // the noumber of many (>=2)-element cycles
    int longestC = 1;     // length of the longest cycle
    bool diffLength = false;
    // if there are cycles > 1 of different length (R+)
    // if R+ -- there may be one-element cycle and other cycles have to have
    // the same length

    for (int lc = 0; lc < noC; lc++)
    {
        int lenCyc = cycles[lc].size();
        if (lenCyc > longestC)
            longestC = lenCyc;

        if (lenCyc == 1)
        {
            oneElementC++;
        }
        else
        {
            if (manyElementC != 0)
            {
                if (lenCyc != longestC)
                    diffLength = true;
            }
            manyElementC++;
        }
    }

    if (sg) // R+ --- rotation
    {
        bool firstO = true;
        cout << "det(R) = 1" << endl;

        if ((oneElementC > 1) || diffLength) // something wrong
        {
            if (oneElementC == noC)
            {
                cout << "Identity." << endl;
            }
            else
            {
                cout << "Ups...!" << endl;
                cout << "There are many one-element cycles or many-element cycles of different length!" << endl;
            }
        }
        else // there is at most one-element cycle and other are of the same length
        {
            for (int lc = 0; lc < noC; lc++)
            {
                int lenCyc = cycles[lc].size();
                if (lenCyc == 1)
                {
                    if (firstO)
                    {
                        cout << "Rotation around point";
                        firstO = false;
                    }
                    cout << "q" << cycles[lc][0];
                }
            }

            cout << endl
                 << "Rotation angle = 2*Pi/" << longestC << endl
                 << endl;
        }
    }
    else // R- --- reflection, trace(R) = 0!
    {
        cout << "det(R) = -1" << endl;

        MyInterval tr = trace(R, 2);
        if (!isSingular(tr) || (longestC > 2))
        // something wrong
        {
            cout << "Ups...!" << endl;
            cout << "trace R = " << tr << "or ";
            cout << "longest cycle = " << longestC << endl;
        }

        int qi = 0;
        int qj = 0;
        for (int lc = 0; lc < noC; lc++)
        {
            int lenCyc = cycles[lc].size();
            if (lenCyc == 2)
            {
                qi = cycles[lc][0];
                qj = cycles[lc][1];
            }
        }
        // reflection with respect to the plane qj - qi
        if (qi != qj)
        {
            cout << "Reflection with respect to the bisecting line of the segment "
                 << "[q" << qj << ", q" << qi << "]" << endl;
        }
        else
        {
            cout << "Identity!" << endl;
        }
    }
} // writesymmetryFromPerm2D

// 2D
int FindAndProveOrthogonalSymmetriesWithBodyA2D(const Bodies &bodies, int bodyA,
                                                ofstream &outFileTest, ofstream &outFileFailed)
// the function constructs all ortogonal symmetries such that N-2 ->  bodyA
//       and does the rigorous verification
//   returns number of ortogonal symmetries different from identity
//
//  in the construction of possible ortogonal transformation we identify bodyA which will be image of 	q_{n-2}
//    this gives us the normalized direction of bodyA - an image of the axis OX
//
//   permutation[N-2] = bodyA;

{

    int numOfSym = 0;
    int N = bodies.N;

    MyInterval xA = bodies.x(bodyA);
    MyInterval yA = bodies.y(bodyA);
    // ex = (dirEXx, dirEXy) --- image of coordinates axis  OX by  R
    MyInterval distxA = normEucl2D(xA, yA);
    MyInterval dirEXx = xA / distxA;
    MyInterval dirEXy = yA / distxA;

    MyInterval dirEYx = -dirEXy;
    MyInterval dirEYy = dirEXx;

    bool RminusOnly = false;
    if (bodyA == N - 2) // R would be identity or reflection y --> -y
    {
        RminusOnly = true; // identity or reflection
    }

    /** matrix R  = [ex| ey|
    * tantatively maps q_{N-2} onto q_{bodyA}
    * RT is inverse of R
    */

    MyMatrix R(2, 2);
    R[0][0] = dirEXx;
    R[0][1] = dirEYx;
    R[1][0] = dirEXy;
    R[1][1] = dirEYy;
    MyMatrix RT(2, 2);
    transposeMatrix(2, R, RT);

    MyMatrix Rminus(2, 2);
    Rminus = R;
    Rminus[0][1] = -dirEYx;
    Rminus[1][1] = -dirEYy;
    MyMatrix RTminus(2, 2);
    transposeMatrix(2, Rminus, RTminus);

    // up to now we construced R and R^{-1}=RT such that
    //  there is possibility that N-2 -> bodyA
    //                             0 -> bodyB

    int *permutation = new int[N];
    if (!RminusOnly)
    {
        if (establishPermutation(bodies, permutation, R))
        {
            if (proveSymmetryWriteToReportFile(bodies, permutation, RT, true, outFileTest, outFileFailed))
            {
                // true - means orientation preserving
                numOfSym++;
            }
        }
    }

    if (establishPermutation(bodies, permutation, Rminus))
    {

        if (proveSymmetryWriteToReportFile(bodies, permutation, RTminus, false, outFileTest, outFileFailed))
        {
            // false - means orientation reversing
            numOfSym++;
        }
    }
    delete[] permutation;
    return numOfSym;
} // FindAndProveOrthogonalSymmetriesWithBodyA2D
