#include "defo.h"
#include "find-symmetry.h"

MyInterval normEucl3D(MyInterval x, MyInterval y, MyInterval z)
/** Returns Euclidean norm of vector (x,y,z) */
{
    return sqrt((power(x, 2) + power(y, 2) + power(z, 2)));
}

void reflectXY(const Bodies &bodies, Bodies &mirrorReflection)
/** Reflection with respect to one the plane OXY
    x->x , y->y, z->-z
*/
{
    MyInterval temp;

    mirrorReflection = bodies;
    for (int i = 0; i < bodies.N - 1; i++)
    {
        temp = -bodies.z(i);
        mirrorReflection.setZ(i, temp);
    }
}

bool flatCC(const Bodies &bodies,
            ofstream &outFileTest, ofstream &outFileFailed)
// returns true   the CC is planar solution (in 3D), i.e. is contained in the plane OXY (this is the only posibility to be planar)
//  uses global variables: zerosTestedFile,  undecidedFile in a non-essential way - which should be
//   removed
{
    for (int i = 0; i < bodies.N; i++)
    {
        if (!isSingular(bodies.z(i)))
            return false;
    }

    Bodies mirrorReflection(bodies);
    Bodies hullBodies(bodies);

    reflectXY(bodies, mirrorReflection);
    intervalHullBodies(bodies, mirrorReflection, hullBodies);

    // the trash variables used for compatybility with the call of searchCC
    //ofstream outFileTest;
    //ofstream outFileFailed;
    //outFileTest.open(zerosTestedFile, ios::binary | ios::out | ios::trunc);
    //outFileFailed.open(undecidedTestFile, ios::binary | ios::out | ios::trunc);
    // end of trash

    // 6.10.2020
    bool doSearch = false;
    bool nonDegeneracyTest = false;
    Count zeroFound = searchCC(hullBodies, doSearch, nonDegeneracyTest, outFileTest, outFileFailed); // testing stage

    //    outFileTest.close();
    //    outFileFailed.close();

    bool isUniqueSol = isUnique(zeroFound);
    if (!isUniqueSol)
        isUniqueSol = blowUpAndProveUniqueSolution(hullBodies, false); // w fazie testowanie nonDegeneracyTest = false
    if (!isUniqueSol)
        cout << "Couldn't establish uniqueness in flatness test!" << endl;

    return isUniqueSol;
}

void writeVec(MyInterval x, MyInterval y, MyInterval z)
{
    cout << "(" << x << ", " << y << ", " << z << ") ";
}

MyInterval distFromLine(MyInterval xk, MyInterval yk, MyInterval zk, MyInterval dirX, MyInterval dirY, MyInterval dirZ)
//  returns: the distance of the point (xk,yk,zk) from the line t*(dirX,dirY,dirZ)
//      it is assumed that |(dirX,dirY,dirZ)|=1
{
    MyInterval sc = scalarProduct(xk, yk, zk, dirX, dirY, dirZ);
    return normEucl3D(xk - sc * dirX, yk - sc * dirY, zk - sc * dirZ);
}

void vectProduct(MyInterval dirEXx, MyInterval dirEXy, MyInterval dirEXz,
                 MyInterval dirEYx, MyInterval dirEYy, MyInterval dirEYz,
                 MyInterval &dirEZx, MyInterval &dirEZy, MyInterval &dirEZz)
// On output:   MyInterval (dirEZx,dirEZy,dirEZz)=(dirEXx,dirEXy,dirEXz) x (dirEYx,dirEYy,dirEYz)

{
    dirEZx = dirEXy * dirEYz - dirEXz * dirEYy;
    dirEZy = dirEXz * dirEYx - dirEXx * dirEYz;
    dirEZz = dirEXx * dirEYy - dirEXy * dirEYx;
}

void candidatesForB(const Bodies &bodies, int bodyA, MyInterval dirx, MyInterval diry, MyInterval dirz,
                    vector<int> &possibleB)
// in the construction of possible ortogonal transformation we first identify bodyA which will be image of 	q_{n-2}
//    this gives us the direction (dirx,diry, dirz) - the normalized direction of bodyA - an image of the axis OX
//  now the bodyB, the image of q_0, is found as the point which is furtherest from this line
/*
 * possibleB[] contains indicies of bodies which might be images of q_0
 */
{
    int N = bodies.N;
    possibleB.clear();
    MyInterval absY0 = abs(bodies.y(0));
    //cout << "absY0" << absY0 << endl;

    for (int k = 0; k < N; k++)
    {
        if (k == bodyA)
            continue;
        MyInterval xk = bodies.x(k);
        MyInterval yk = bodies.y(k);
        MyInterval zk = bodies.z(k);
        MyInterval tk = distFromLine(xk, yk, zk, dirx, diry, dirz);

        if (!emptyIntersection(tk, absY0))
        {
            possibleB.push_back(k);
        }
    }
}

// 3D
void writesymmetryFromPerm3D(int permutation[], int N, vector<vector<int>> &cycles, int noC, bool sg, MyMatrix R)
// the function writes to cout the data about symmetry
{
    // sg --- sign of det of matrix R:
    // sg = TRUE <==> det(R) = 1 <==> R+
    // sg = FALSE <==> det(R) = -1 <==> R-
    // noC -- number of cycles in vector 'cycles'
    int oneElementC = 0;  // the noumber of 1-element cycles
    int twoElementC = 0;  // the noumber of 2-element cycles
    int manyElementC = 0; // the noumber of many (>2)-element cycles
    int longestC = 1;     // length of the longest cycle

    for (int lc = 0; lc < noC; lc++)
    {
        int lenCyc = cycles[lc].size();
        if (lenCyc > longestC)
            longestC = lenCyc;

        if (lenCyc == 1)
        {
            oneElementC++;
        }
        else if (lenCyc == 2)
        {
            twoElementC++;
        }
        else
        {
            manyElementC++;
        }
    }

    if (sg) // R+
    {
        bool firstO = true;
        cout << "det(R) = 1" << endl;
        for (int lc = 0; lc < noC; lc++)
        {
            int lenCyc = cycles[lc].size();
            if (lenCyc == 1)
            {
                if (firstO)
                {
                    cout << "Rotation axis: 0";
                    firstO = false;
                }
                cout << "-q" << cycles[lc][0];
            }
        }

        cout << endl
             << "Rotation angle = 2*Pi/" << longestC << endl
             << endl;
    }
    else // R-
    {
        cout << "det(R) = -1" << endl;

        MyInterval tr = trace(R, 3);
        //cout << "trace R = " << tr << endl;
        if (manyElementC == 0) // either (-Id) or reflection with respect to some plane
        {
            if (twoElementC == 0)
                cout << "Reflection with respect to the plane Z = 0" << endl;
            else
            {
                if (tr < 1)
                {
                    cout << "-Identity (-Id)" << endl;
                }
                else if (tr > -3)
                {
                    int qi = 0;
                    int qj = 0;
                    for (int lc = 0; lc < noC; lc++)
                    {
                        int lenCyc = cycles[lc].size();
                        if (lenCyc == 2)
                        {
                            qi = cycles[lc][0];
                            qj = cycles[lc][1];
                        }
                    }
                    // reflection with respect to the plane qj - qi
                    cout << "Reflection with respect to the bisecting plane of the segment "
                         << "[q" << qj << ", q" << qi << "]" << endl;
                }
                else // it should never happen!!!
                {
                    cout << "We cannot establish a kind of symmetry!" << endl;
                }
            }
        }
        else
        {
            MyInterval cosPhi = (tr + 1) / 2.0;
            MyInterval posCos1 = cos(2 * piConst / longestC);
            MyInterval posCos2 = cos(4 * piConst / longestC);
            //cout << "cosPhi = " << cosPhi << ", posCos1 = " << posCos1 << ", posCos2 = " << posCos2 << endl;
            if (!emptyIntersection(cosPhi, posCos1) && emptyIntersection(cosPhi, posCos2))
            {
                cout << "Rotation angle = 2*Pi/" << longestC << endl
                     << endl;
            }
            else if (emptyIntersection(cosPhi, posCos1) && !emptyIntersection(cosPhi, posCos2))
            {
                cout << "Rotation angle = Pi/" << longestC << endl
                     << endl;
            }
            else if (!emptyIntersection(cosPhi, posCos1) && !emptyIntersection(cosPhi, posCos2))
            { // it should never happen!!!
                cout << "Rotation angle = 2*Pi/" << longestC << " or Pi/" << longestC << endl
                     << endl;
            }
            else // it should never happen!!!
            {
                cout << "An error: in computation of rotation angle!" << endl;
            }
        }
    }
}

// 3D
int FindAndProveOrthogonalSymmetriesWithBodyA3D(const Bodies &bodies, int bodyA,
                                                ofstream &outFileTest, ofstream &outFileFailed)
// the function constructs all ortogonal symmetries such that N-2 ->  bodyA
//       and does the rigorous verification
//   returns number of ortogonal symmetries different from identity
//
//  in the construction of possible ortogonal transformation we first identify bodyA which will be image of 	q_{n-2}
//    this gives us the direction (dirx,diry, dirz) - the normalized direction of bodyA - an image of the axis OX
//  now the bodyB, the image of q_0, is found as the point which is furtherest from this line
/**
 *   permutation[N-2] = bodyA;
 *   permutation[0] = bodyB;
     this fixes the image of the plane OXY
	 and then it remains the decide what will be the image of OZ-axis , we have two possibilities the orientation preserving and non-preserving
 *
 *
 */
{

    int numOfSym = 0;
    int N = bodies.N;

    MyInterval xA = bodies.x(bodyA);
    MyInterval yA = bodies.y(bodyA);
    MyInterval zA = bodies.z(bodyA);
    // ex = (dirEXx, dirEXy, dirEXz) --- image of coordinates axis  OX by  R
    MyInterval distxA = normEucl3D(xA, yA, zA);
    MyInterval dirEXx = xA / distxA;
    MyInterval dirEXy = yA / distxA;
    MyInterval dirEXz = zA / distxA;

    vector<int> possibleB;

    candidatesForB(bodies, bodyA, dirEXx, dirEXy, dirEXz, possibleB);

    bool RminusOnly;
    for (auto &&bodyB : possibleB)
    {

        RminusOnly = false;
        if (bodyA == N - 2 && bodyB == 0) // R would be identity or reflection z --> -z
        {
            RminusOnly = true; // the orientation preserving is the identity
        }

        // we construct ey = (dirEYx, dirEYy, dirEYz)
        MyInterval xB = bodies.x(bodyB);
        MyInterval yB = bodies.y(bodyB);
        MyInterval zB = bodies.z(bodyB);

        MyInterval sc = scalarProduct(xB, yB, zB, dirEXx, dirEXy, dirEXz);
        MyInterval tildeX = xB - sc * dirEXx;
        MyInterval tildeY = yB - sc * dirEXy;
        MyInterval tildeZ = zB - sc * dirEXz;
        MyInterval normTildeEy = normEucl3D(tildeX, tildeY, tildeZ);
        //cout << "normTildeEy = " << normTildeEy << endl;

        if (isSingular(normTildeEy))
        {
            cout << "Cannot normalize ey vector --- likely collinear vectors! " << endl;
            return false;
        }
        MyInterval dirEYx = (tildeX) / normTildeEy;
        MyInterval dirEYy = (tildeY) / normTildeEy;
        MyInterval dirEYz = (tildeZ) / normTildeEy;

        // ez = ex X ey
        MyInterval dirEZx;
        MyInterval dirEZy;
        MyInterval dirEZz;
        vectProduct(dirEXx, dirEXy, dirEXz, dirEYx, dirEYy, dirEYz, dirEZx, dirEZy, dirEZz);
        MyInterval normEz = normEucl3D(dirEZx, dirEZy, dirEZz);
        if (isSingular(normEz))
        {
            cout << "Cannont normalize ez vector! " << endl;
            return false;
        }

        /** matrix R  = [ex| ey| +- ez]
    * tantatively maps q_{N-2} onto q_{bodyA}
    * q_0 onto q_{bodyB}
    * RT is inverse of R
    */

        MyMatrix R(3, 3);
        ///*
        R[0][0] = dirEXx;
        R[0][1] = dirEYx;
        R[0][2] = dirEZx;
        R[1][0] = dirEXy;
        R[1][1] = dirEYy;
        R[1][2] = dirEZy;
        R[2][0] = dirEXz;
        R[2][1] = dirEYz;
        R[2][2] = dirEZz;
        //*/
        MyMatrix RT(3, 3);
        transposeMatrix(3, R, RT);

        MyMatrix Rminus(3, 3);
        Rminus = R;
        Rminus[0][2] = -dirEZx;
        Rminus[1][2] = -dirEZy;
        Rminus[2][2] = -dirEZz;
        MyMatrix RTminus(3, 3);
        transposeMatrix(3, Rminus, RTminus);

        // up to now we construced R and R^{-1}=RT such that
        //  there is possibility that N-2 -> bodyA
        //                             0 -> bodyB

        int *permutation = new int[N];
        if (!RminusOnly)
        {
            if (establishPermutation(bodies, permutation, R))
            {
                if (proveSymmetryWriteToReportFile(bodies, permutation, RT, true, outFileTest, outFileFailed))
                {
                    // true - means orientation preserving
                    numOfSym++;
                }
            }
        }

        if (establishPermutation(bodies, permutation, Rminus))
        {

            if (proveSymmetryWriteToReportFile(bodies, permutation, RTminus, false, outFileTest, outFileFailed))
            {
                // false - means orientation reversing
                numOfSym++;
            }
        }
        delete[] permutation;
    } // for --- all possible coorB
    return numOfSym;
}
