#include <ctime>
#include <iomanip>

// to print current date and time of finding zeros
#include <time.h>
//////////////////////////////////////////////////

#include "defo.h"

#include "bodies.h"
#include "krawczyk-newton.h"
#include "my-interval-func.h"
#include "test-and-help-functions.h"
#include "search-alg.h"

#include "cc-nb.h"

#ifdef TESTY
#include "testy.h"
#endif

void modifyPath(vector<bool> path, int dir, vector<bool> &pathLocal)
{

    pathUpdate_mutex.lock();
    pathLocal.clear();
    // A loop to copy elements of path  into a vector pathLocal
    for (unsigned i = 0; i < path.size(); i++)
        pathLocal.push_back(path[i]);

    pathLocal.push_back(dir); // dir -- direction in the computation tree: 0 -- to indicate we take left branch, 1 --- right
    pathUpdate_mutex.unlock();
}

void writePath(Bodies bodies, const vector<bool> &path, Count result)
{
    // best parameters to see what happens...
    // -------- 2D ------------
    // N     pathLength    printData
    // 4      60              63
    // 5      70              74
    // -------- 3D ------------
    // 4      70              72
    // 5     103             105
    pathUpdate_mutex.lock();
    long pathLength = path.size();
    /*
    if (result.zeros == 1)
    {
        cout << "Program finished path: ";
        for (int i = 0; i < path.size(); i++)
            cout << path[i] << " ";
        cout << endl;
        cout << "and found CC:" << endl;
        bodies.printData();
        cout << "Path's length = " << pathLength << endl;
        cout << "- - - - - - - - - - - - - - - -- - - - - " << endl;
        cout << endl;
    }
    */

    if (pathLength >= 70)
    {
        cout << "Program finished path: ";
        for (unsigned i = 0; i < path.size(); i++)
            cout << path[i] << " ";
        cout << endl;

        if (pathLength >= 72)
        {
            if (result.zeros == 0)
            {
                cout << "There is no CCs" << endl;
                cout << "Path's length = " << pathLength << endl;
                cout << "- - - - - - - - - - - - - - - -- - - - - " << endl;
                cout << endl;
            }
            else if (result.undecided != 0)
            {
                cout << "Undecided box" << endl;
                cout << "Path's length = " << pathLength << endl;
                cout << "- - - - - - - - - - - - - - - -- - - - - " << endl;
                cout << endl;
            }
            else
            {
                cout << "Deepest box:" << endl;
                bodies.printData();
                cout << "Path's length = " << pathLength << endl;
                cout << "- - - - - - - - - - - - - - - -- - - - - " << endl;
                cout << endl;
            }
        }
    }
    pathUpdate_mutex.unlock();
}

/**
 * The asynchronous search:
 * The search fires recursively on both parts of the subdivided interval
 * unless the stopping conditions are met. Each recursive call tracks the
 * adds one to the global count of parallel searchers running. We don't
 * perform the call asynchronously if certain threshold has been reached.
 * This is because we want to avoid the overhead of starting to many threads
 * on the host system. Note that once the asynchronous search function finishes,
 * it adds back to the pool of available threads, so later subdvisions can again
 * be performed asynchronously.
 */
#ifdef PATH_SHOW
Count searchCC(Bodies bodies, bool searchStage, bool nonDegeneracyTest, ostream &fhsolfound, ostream &fhboxfailed, vector<bool> path)
#endif

#ifndef PATH_SHOW
    Count searchCC(Bodies bodies, bool searchStage, bool nonDegeneracyTest, ostream &fhsolfound, ostream &fhboxfailed)
#endif
//  solutions (boxes) found are written to   <fhsolfound>, the undecided boxes to <fhboxfailed> in binary format
{
    if (searchStage)
        boxCount++; // the global variable counting all boxes treated

    ResultType newtonRes = ResultType::uniqueZero;
    Count result = {0, 0};
    Count leftCount, rightCount;
    std::future<Count> leftCountF, rightCountF;

    // 6.10.2020
    // init returned variable "result"
    result.zeros = 0;
    result.undecided = 0;
    result.nonDegenerate = 0;

    if (searchStage)
    {

#ifdef TESTY
        if (bodiesEmptyIntersection(bodies, *testBodies))
        {
            // 6.10.2020
            //result.zeros = 0;
            //result.maybeZeros = 0;
            return result;
        }
#endif
        if (!possibleSolution(bodies)) // bodies - can be updated (shrinked) in the execution of this function
                                       // so that all possible CC in bodies (on input) are contained in bodies(on output)
        {
            // 6.10.2020
            //result.zeros = 0;
            //result.maybeZeros = 0;
#ifdef PATH_SHOW
            writePath(bodies, path, result);
#endif
            return result;
        }
    }

    bool collision = bodies.collisionHappened();

    // We're looking for the longest interval to subdivide.
    // The found interval describes the max_it-th body.
    // coord denotes whether to subdivide on x, y or z coord.
    int max_i = 0;
    int coord = 0; // coord = 0 => subdivide x, coord = 1 => subdivide y, coord = 2 => subdivide z
    MyInterval maxLengthInterval = bodies.maxDiam(max_i, coord);
    bool DFFComputed = false;

    // do celow diagniostycznych
    // int min_i = 0;
    // int coordMin = 0; // coord = 0 => subdivide x, coord = 1 => subdivide y, coord = 2 => subdivide z
    // MyInterval minLengthInterval = bodies.minDiam(min_i, coordMin);
    //

    bool intersec = false; // when true then we know that we got an improvement in Newton/Krawczyk method
    int dimRCS = bodies.dimRedConfSpace();
    MyMatrix DFF(dimRCS, dimRCS);

    if (!collision)
    {
        if (maxLengthInterval < bias) // bias is a global variable, we call the costly Nowton/Krawczyk method, when  the size of the set is <= eps (another global variable)
        {
            // possible zero;
            // use Newton's/Krawczyk's method to check if that's true:
            Bodies bodiesRetFromKraw(bodies.N);
            //            int dimRCS = bodies.dimRedConfSpace();
            //            MyMatrix DFF(dimRCS, dimRCS);
            computeDFRedMatrix(bodies, DFF);
            DFFComputed = true;

            // 6.01.2020
            if (nonDegeneracyTest)
                if (isomorphismTest(DFF))
                {
                    // if the program is run in the mode of testing nonDegeneracyTest of the number of CC's,
                    // then zeros and undecided fields are neglected
                    // we use nonDegenerate to count the  upper bound of the number of CC's
                    result.nonDegenerate = 1;
                    return result;
                }
            bool recomputeDFF=false;
			#ifdef   RECOMPUTE_DFF_krawczyk
			recomputeDFF=true;
 			#endif
			bool KrawczykAsRefinement=false;
            newtonRes = (methodUsed == ZeroSearchType::newton) ? newtonMethod(bodies, bodiesRetFromKraw, intersec, DFF) : 
			                                krawczykMethod(bodies, bodiesRetFromKraw, intersec, DFF, recomputeDFF,KrawczykAsRefinement);
            // if intersec=true, then we know that we got an improvement (bodies was shrinked) in Newton/Krawczyk method
            // HOWEVER it might happen that box was shrinked and  now is extremely small but still undecided
            // first we try to blowUp --- if it hepls we have unique solution in a box
            // if not --- we return undecide box, but we reutrn bodies that enter Krawczyk method --- the whole box is undecided
            maxLengthInterval = bodiesRetFromKraw.maxDiam(max_i, coord); // we update max_i and coord because <bodies> has been changed

            switch (newtonRes)
            {
            case ResultType::methodFailed:
                if (searchStage)
                    KrawczykFailedCounter++;
                // 6.10.2020
                if (maxLengthInterval >= eps) // box is big enough for further computations
                {
                    bodies = bodiesRetFromKraw; // we gained something, the size is smaller
                }
                break;

            case ResultType::uniqueZero:
                if (searchStage)
                {
                    KrawczykZeroCounter++;

                    if (phase == searchmode)
                    {
                        // time stamp -- BEGIN
                        cout << "------------------------ zero no " << KrawczykZeroCounter << " ------------------------" << endl;
                        // declaring argument of time()
                        time_t my_time = time(NULL);

                        // ctime() used to give the present time
                        printf("%s", ctime(&my_time));
                        bodiesRetFromKraw.printData(); // time stamp -- END
                    }
                }
                //methodZeroes++;
                // 6.10.2020
                //if (!nonDegeneracyTest)
                {
                    krawczykSection_mutex.lock();
                    bodiesRetFromKraw.writeBinaryData(fhsolfound); //
                    krawczykSection_mutex.unlock();
                }

                result.zeros = 1;
#ifdef PATH_SHOW
                writePath(bodies, path, result);
#endif
                return result;

            case ResultType::noZeroInSet:
                if (searchStage)
                {
                    KrawczykNoZeroCounter++;
#ifdef PATH_SHOW
                    writePath(bodies, path, result);
#endif
                }
                return result;

            default:
                break;
            }
        }
    }

    if (maxLengthInterval < eps)
    // if the size of the cube is too small, then we interrupt search
    {
        if (searchStage)
        {
            methodFailedCounter++; // global variable counts undecided cubes!
            // 6.10.2020
            {
                write_mutex.lock();
                bodies.writeBinaryData(fhboxfailed); // !
                write_mutex.unlock();

#ifdef PATH_SHOW
                writePath(bodies, path, result);
#endif
            }
        }
        result.undecided = 1;

        return result;
    }

    /* 6.10.2020
    if (DFFComputed)
    {
        int max_iDFF;
        int coordDFF;
        fixDirection(bodies, DFF, dimRCS, max_iDFF, coordDFF);
        // wybieramy kierunek dfdx jesli pochodne nie sa zbyt duze
        int min_i = 0;
        int coordMin = 0; // coord = 0 => subdivide x, coord = 1 => subdivide y, coord = 2 => subdivide z
        MyInterval minLengthInterval = bodies.minDiam(min_i, coordMin);
        if (minLengthInterval > eps) // jezeli najkrotszy przedzial jest wciaz dlui
        // to mozemy spokojnie dzielic
        {
            max_i = max_iDFF;
            coord = coordDFF;
        }
    }
    */

    // 6.10.2020
    // set max_i, coord
    midPointInt(bodies.maxDiam(max_i, coord));

    // we do binary division in direction of max_i coordinate
    // the length of diveded interval is maxLengthInterval
    MyDouble leftPoint = 0, rightPoint = 0, midPoint = 0, leftMidPoint = 0, rightMidPoint = 0;
    MyDouble overlap = capd::min(0.1 * eps, 1e-6); // overlap has to be less than 0.5
    if (coord == 0)                                // x-variable
    {
        leftPoint = bodies.x(max_i).leftBound() - overlap;
        rightPoint = bodies.x(max_i).rightBound() + overlap;
    }
    if (coord == 1) // y-variable
    {
        leftPoint = bodies.y(max_i).leftBound() - overlap;
        rightPoint = bodies.y(max_i).rightBound() + overlap;
    }
#ifdef SPATIAL
    if (coord == 2) // z-variable
    {
        leftPoint = bodies.z(max_i).leftBound() - overlap;
        rightPoint = bodies.z(max_i).rightBound() + overlap;
    }
#endif
// 6.10.2020
#ifdef MASSDIV
    if (coord == 3) // mass-variable
    {
        // 6.10.2020 -- jesli dziele w przestrzeni mas, to nie uzywam overlap, ale po co go zerowac?
        // overlap = 0.0; // in mass space there is no need for overlap
        leftPoint = bodies.mass(max_i).leftBound();
        rightPoint = bodies.mass(max_i).rightBound();
    }
#endif

    midPoint = (leftPoint + rightPoint) / 2.0;
    rightMidPoint = midPoint + (rightPoint - leftPoint) * overlap;
    leftMidPoint = midPoint - (rightPoint - leftPoint) * overlap;

    // launching the computation on the "left-half" of bodies
    if (coord == 0) // x-variable
        bodies.setX(max_i, MyInterval(leftPoint, rightMidPoint));
    if (coord == 1) // y-variable
        bodies.setY(max_i, MyInterval(leftPoint, rightMidPoint));
#ifdef SPATIAL
    if (coord == 2) // z-variable
        bodies.setZ(max_i, MyInterval(leftPoint, rightMidPoint));
#endif
#ifdef MASSDIV
    if (coord == 3)
    { // mass-variable
        bodies.setMass(max_i, MyInterval(leftPoint, rightMidPoint));
    }
#endif

    bool shouldAsync = cnt < MAX_ASYNC_DEPTH;
#ifdef PATH_SHOW
    // Declaring new vector for path
    vector<bool> pathLocal = {};
#endif

    auto policy = std::launch::async; //| std::launch::deferred;
    if (shouldAsync)
    {
        cnt += 1; // cnt*2 --- threads' counter;  we will launch two new searchCC for the left and right half

#ifdef PATH_SHOW
        if (searchStage)
            modifyPath(path, 0, pathLocal);
        leftCountF = std::async(policy, [bodies, searchStage, nonDegeneracyTest, &fhsolfound, &fhboxfailed, pathLocal]() { return searchCC(bodies, searchStage, nonDegeneracyTest, fhsolfound, fhboxfailed, pathLocal); });
#endif
#ifndef PATH_SHOW
        leftCountF = std::async(policy, [bodies, searchStage, nonDegeneracyTest, &fhsolfound, &fhboxfailed]() { return searchCC(bodies, searchStage, nonDegeneracyTest, fhsolfound, fhboxfailed); });
#endif
    }
    else
    {
#ifdef PATH_SHOW
        if (searchStage)
            modifyPath(path, 0, pathLocal);
        leftCount = searchCC(bodies, searchStage, nonDegeneracyTest, fhsolfound, fhboxfailed, pathLocal);
#endif
#ifndef PATH_SHOW
        leftCount = searchCC(bodies, searchStage, nonDegeneracyTest, fhsolfound, fhboxfailed);
#endif
    }

    // now on the "right-half"
    if (coord == 0) // x-variable
        bodies.setX(max_i, MyInterval(leftMidPoint, rightPoint));
    if (coord == 1) // y-variable
        bodies.setY(max_i, MyInterval(leftMidPoint, rightPoint));
#ifdef SPATIAL
    if (coord == 2) // z-variable
        bodies.setZ(max_i, MyInterval(leftMidPoint, rightPoint));
#endif
#ifdef MASSDIV
    if (coord == 3)
    {
        // mass-variable
        bodies.setMass(max_i, MyInterval(leftMidPoint, rightPoint));
    }
#endif

    if (shouldAsync)
    {

#ifdef PATH_SHOW
        if (searchStage)
            modifyPath(path, 1, pathLocal);
        rightCountF = std::async(policy, [bodies, searchStage, nonDegeneracyTest, &fhsolfound, &fhboxfailed, pathLocal]() { return searchCC(bodies, searchStage, nonDegeneracyTest, fhsolfound, fhboxfailed, pathLocal); });
#endif
#ifndef PATH_SHOW
        rightCountF = std::async(policy, [bodies, searchStage, nonDegeneracyTest, &fhsolfound, &fhboxfailed]() { return searchCC(bodies, searchStage, nonDegeneracyTest, fhsolfound, fhboxfailed); });
#endif

        leftCount = leftCountF.get();
        rightCount = rightCountF.get();
        cnt -= 1; // two synchrounpus searchCC has been completed
    }
    else
    {
#ifdef PATH_SHOW
        if (searchStage)
            modifyPath(path, 1, pathLocal);
        rightCount = searchCC(bodies, searchStage, nonDegeneracyTest, fhsolfound, fhboxfailed, pathLocal);
#endif
#ifndef PATH_SHOW
        rightCount = searchCC(bodies, searchStage, nonDegeneracyTest, fhsolfound, fhboxfailed);
#endif
    }
    /*
    write_mutex.lock();
    if (minLengthInterval.leftBound() >= 0.85 && minLengthInterval.leftBound() < 1.0)
    {
        cout << "minLengthInterval = " << minLengthInterval << endl;
        std::time_t t = std::time(nullptr);
        std::cout << std::ctime(&t);
    }
    else if (minLengthInterval.leftBound() >= 0.75 && minLengthInterval.leftBound() < 0.85)
    {
        cout << endl;
        std::time_t t = std::time(nullptr);
        std::cout << std::ctime(&t);
        cout << "minLengthInterval = " << minLengthInterval << endl;
        bodies.printData();
        cout << "----------------------------------------------------------------------------------------" << endl;
    }
    write_mutex.unlock();
    */
    // 6.10.2020
    result.zeros = leftCount.zeros + rightCount.zeros;
    result.undecided = leftCount.undecided + rightCount.undecided;
    result.nonDegenerate = leftCount.nonDegenerate + rightCount.nonDegenerate;

#ifdef PATH_SHOW
    if (searchStage)
        writePath(bodies, path, result);
#endif

    return result;

} //search

// 6.10.2020
bool blowUpAndProveUniqueSolution(Bodies &bodies, bool nonDegeneracyTest)
/* to be used on  <bodies> with low diameter;
  we want to prove that <bodies> contain a unique solution;

  this function is used during testing uniqueness and symmetry
//  returns true, when the Krawczyk/Newton method have shown that there is indeed unique solution in <bodies>
//     and false otherwise
*/
///* wersja 23.05.2020 -- poczatek
{
    int ret = blowUpSearch(bodies, nonDegeneracyTest);
    if (ret == 1)
        return true; // there is unique solution in puffedBodies
    else if (ret == 3)
    {
        cout << "w blowUp w fazie testowania pojawilo sie wiecej zer!" << endl;
        return false;
    }
    else
        return false;
}
//*/ //wersja 23.05.2020 -- koniec

int blowUpSearch(Bodies &bodies, bool nonDegeneracyTest)
// this function is used during testing undecided boxes and
// (in blowUpAndProveUniqueSolution) during testing uniqueness and symmetry
/* to be used on  <bodies> with low diameter;
  we want to prove that <bodies> contain a unique solution;

  sometimes in interval arithmetic can happen that on the result from
  the Krawczyk operator we cannot establish the existence, then we should apply this function
  for example when the solution was obtained in more optimised calculations

  on output:
  0 -- there is no solution
  1 -- there is unique solution
  2 -- puffedBodies remains undecided
  3 -- there are more than 1 solutions (all solutions proved by Krawczyk); we do not know how many undecided boxes are
  */

{
    /*
    in the loop we "puff" the boides and try to (dis)prove the existence of the solution in the set
     if not, the zero can still be present
   */

    //  now we start puffing the bodies
    Bodies *midBodies = bodies.calculateMidBodies();
    int trash1 = 0, trash2 = 0;
    MyInterval maxLengthInterval = bodies.maxDiam(trash1, trash2);
    MyInterval half = MyInterval(maxLengthInterval) / 2.0;
    half = MyInterval(-half.rightBound(), half.rightBound());
    Bodies puffedBodies(bodies);
    // it is important that puffedBodies are created as copy of <bodies>, bacause in this way the masses are copied and
    // maybe some other setting introduced in the future
    calculatePuffedBodies(*midBodies, half, puffedBodies); // at the moment bodies \subset puffedBodies
    delete midBodies;

    int whatSearchDecides;
    bool theEnd = false;
    while (!theEnd)
    {

        Count result = {0, 0, 0};
        // we define some trash variables for compatybility with the call of searchCC
        ofstream trashFileZeros;
        ofstream trashFileFailed;
        trashFileZeros.open("data/trash-zeros.dat", ios::binary | ios::out | ios::trunc);
        trashFileFailed.open("data/trash-undecided.dat", ios::binary | ios::out | ios::trunc);
        // trash-end

        bool searchStage = true; // we want to find CCs in puffedBodies
#ifdef PATH_SHOW
        const std::vector<bool> path; // a path from root to leaves in the computation tree
        result = searchCC(puffedBodies, searchStage, nonDegeneracyTest, trashFileZeros, trashFileFailed, path);
#endif
#ifndef PATH_SHOW
        result = searchCC(puffedBodies, searchStage, nonDegeneracyTest, trashFileZeros, trashFileFailed);
#endif
        trashFileZeros.close();
        trashFileFailed.close();

        /*
        cout << "nondegeneracy = " << result.nonDegenerate << endl;
        cout << "zeros = " << result.zeros << endl;
        cout << "undecided = " << result.undecided << endl
             << endl
             << flush;
*/
        if (result.undecided == 0)
        {
            if (result.zeros == 0)
                whatSearchDecides = 0; // there is no solution in puffedBodies
            else
            {
                if (result.zeros == 1)
                {
                    bodies = puffedBodies;
                    whatSearchDecides = 1; // there is unique solution in puffedBodies
                }
                else
                {
                    whatSearchDecides = 3; //non unique zeros!
                }
            }
        }
        else //(result.undecided > 0)
        {
            if (result.zeros == 0)
                whatSearchDecides = 2; // puffedBodies remains undecided
            else
                whatSearchDecides = 3; // somthing wrong(!) result.zeros > 0!!!
        }

        switch (whatSearchDecides)
        {
        case 0: // there is no solution in puffedBodies
            theEnd = true;
            break;

        case 1: // there is unique solution in puffedBodies
            theEnd = true;
            break;

        case 2: // puffedBodies remains undecided --- we puffed again
            half = half * 1.2;
            if (half.rightBound() < 1e-6)
            {
#ifdef CC_SYMMETRY_DEBUG_HELP
                cout << "puffedBodies remains undecided --- we puffed again..." << endl;
#endif
                calculatePuffedBodies(puffedBodies, half, puffedBodies);
            }
            else
            { // we puffed a lot without any result
                theEnd = true;
            }
            break;

        case 3: // somthing wrong(!) result.zeros > 0!!!
            theEnd = true;
            break;

        default:
            break;
        }
    }
    cout << endl
         << flush;

    return whatSearchDecides;
}
